# Copyright (c) Mike Jarvis and other collaborators
# See https://github.com/rmjarvis/MockMPI/LICENSE for license information.

from ._version import __version__, __version_info__
from .comm import MockComm
from .exec import mock_mpiexec
