# Copyright (c) Mike Jarvis and other collaborators
# See https://github.com/rmjarvis/MockMPI/LICENSE for license information.

__version__ = '0.8.2'
__version_info__ = tuple(map(int, __version__.split('.')))

